@@ -4,13 +4,15 @@
 Control flow graph for Python programs.
 """
 
-from typing import Any, Tuple, List, Optional, Iterator, Set
+from typing import Any, Deque, Tuple, List, Optional, Iterator, Set, Dict
 import ast  # type: ignore
 from _ast import Compare  # For type-hinting
 import astor  # type: ignore
 import graphviz as gv  # type: ignore
 from graphviz.dot import Digraph  # type: ignore
 import re
+from collections import Counter, deque
+import os
 
 
 class Block(object):
@@ -22,14 +24,16 @@ class Block(object):
     a list of Links that represent control flow jumps.
     """
 
-    __slots__ = [
+    __slots__ = (
         "id",
         "statements",
         "func_calls",
         "predecessors",
         "exits",
         "func_blocks",
-    ]
+        "highlight",
+        "outline",
+    )
 
     def __init__(self, id: int) -> None:
         # Id of the block.
@@ -45,6 +49,10 @@ class Block(object):
         self.exits: List[Link] = []
         # Function blocks within this block
         self.func_blocks: List[FuncBlock] = []
+        # Interactive styling
+        self.highlight: bool = False
+        # Debugger at this block
+        self.outline: bool = False
 
     def __str__(self) -> str:
         if self.statements:
@@ -59,13 +67,21 @@ class Block(object):
             txt += "]"
         return txt
 
-    def at(self) -> Optional[int]:
+    def at(self) -> int:
         """
         Get the line number of the first statement of the block in the program.
         """
         if self.statements and self.statements[0].lineno >= 0:
             return self.statements[0].lineno
-        return None
+        return -1
+
+    def end(self) -> int:
+        """
+        Get the line number of the last statement of the block in the program.
+        """
+        if self.statements and self.statements[-1].lineno >= 0:
+            return self.statements[-1].lineno
+        return -1
 
     def is_empty(self) -> bool:
         """
@@ -74,7 +90,7 @@ class Block(object):
         Returns:
             A boolean indicating if the block is empty (True) or not (False).
         """
-        return len(self.statements) == 0
+        return not self.statements
 
     def get_source(self) -> str:
         """
@@ -114,11 +130,19 @@ class Block(object):
         default = ast.AST if default is None else default
         return type(self.statements[0]) if self.statements else default
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         # Allows Block objects to be used as dict keys and set elements
         # return id(self)
         return self.id
 
+    def add_statement(self, node: ast.AST):
+        self.statements.append(node)
+
+    def add_exit(self, next, exitcase=None) -> None:
+        link = Link(self, next, exitcase)
+        self.exits.append(link)
+        next.predecessors.append(link)
+
 
 class Link(object):
     """
@@ -129,13 +153,15 @@ class Link(object):
     control jump is made.
     """
 
-    __slots__ = ["source", "target", "exitcase"]
+    __slots__ = (
+        "source",
+        "target",
+        "exitcase",
+        "highlight",
+    )
 
     def __init__(
-        self,
-        source: Block,
-        target: Block,
-        exitcase: Optional[ast.Compare] = None,
+        self, source: Block, target: Block, exitcase: Optional[ast.Compare] = None
     ) -> None:
         assert isinstance(target, Block), "Source of a link must be a block"
         assert isinstance(target, Block), "Target of a link must be a block"
@@ -145,6 +171,8 @@ class Link(object):
         self.target = target
         # 'Case' leading to a control flow jump through this link.
         self.exitcase = exitcase
+        # Interactive styling
+        self.highlight: bool = False
 
     def __str__(self) -> str:
         return f"link from {self.source} to {self.target}"
@@ -156,6 +184,14 @@ class Link(object):
             return f"{self}, with exitcase {ast.dump(self.exitcase)}"
         return str(self)
 
+    def jumpfrom(self) -> Optional[int]:
+        """Return the line of source end"""
+        return self.source.end()
+
+    def jumpto(self) -> Optional[int]:
+        """Return the line of target start"""
+        return self.target.at()
+
     def get_exitcase(self) -> str:
         """
         Get a string containing the Python source code corresponding to the
@@ -170,10 +206,43 @@ class Link(object):
 
 
 class FuncBlock(Block):
+    __slots__ = ("args", "name")
+
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.args: List[FuncBlock] = []
-        self.name: str = None
+        self.name: Optional[str] = None
+
+
+class TryBlock(Block):
+    __slots__ = ("except_blocks",)
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.except_blocks: Dict[Optional[Exception], Block] = {}
+
+    def get_source(self) -> str:
+        """
+        Get a string containing the Python source code corresponding to the
+        statements in the block.
+
+        Returns:
+            A string containing the source code of the statements.
+        """
+        if not self.statements[1:]:
+            return "try:"
+        src = ""
+        for statement in self.statements[1:]:  # We just want try body
+            if type(statement) in [ast.If, ast.For, ast.While]:
+                src += astor.to_source(statement).split("\n")[0] + "\n"
+            elif (
+                type(statement) == ast.FunctionDef
+                or type(statement) == ast.AsyncFunctionDef
+            ):
+                src += (astor.to_source(statement)).split("\n")[0] + "...\n"
+            else:
+                src += astor.to_source(statement)
+        return src
 
 
 class CFG(object):
@@ -196,10 +265,11 @@ class CFG(object):
         ast.While: ("hexagon", "#FFBE52"),  # Pale Orange
         ast.Call: ("tab", "#E552FF"),  # Pale Purple
         ast.Return: ("parallelogram", "#98fb98"),  # Pale Green
+        ast.Try: ("Mdiamond", "orange"),
+        ast.Raise: ("house", "#98fb98"),
     }
     DEFAULT = node_styles["default"]
 
-    
     def __init__(self, name: str, asynchr: bool = False, short: bool = True) -> None:
         assert isinstance(name, str), "Name of a CFG must be a string"
         assert isinstance(asynchr, bool), "Async must be a boolean value"
@@ -217,6 +287,9 @@ class CFG(object):
         # Sub-CFGs
         self.classcfgs: dict = {}
         self.isShort = short
+        self.lineno: int = 0
+        self.end_lineno: int = 0
+
 
     def __str__(self) -> str:
         return "CFG for {}".format(self.name)
@@ -235,11 +308,7 @@ class CFG(object):
                     if isinstance(typeobj, type)
                     else str(typeobj)
                 ),
-                _attributes={
-                    "shape": shape,
-                    "style": "filled",
-                    "fillcolor": color,
-                },
+                _attributes={"shape": shape, "style": "filled", "fillcolor": color},
             )
 
         # Adding edge helps with alignment.
@@ -247,6 +316,7 @@ class CFG(object):
         key_subgraph.edge("input", "call", _attributes={"style": "invis"})
         key_subgraph.edge("for", "return", _attributes={"style": "invis"})
         key_subgraph.edge("return", "default", _attributes={"style": "invis"})
+        key_subgraph.edge("try", "raise", _attributes={"style": "invis"})
         return key_subgraph
 
     # default handler for graph nodes
@@ -267,11 +337,7 @@ class CFG(object):
         # simply defining a method `node_ASTNAME`
         typeobject = block.type(default)
         name = typeobject.__name__ if isinstance(typeobject, type) else typeobject
-        style = getattr(
-            self,
-            f"node_{name}",
-            self._style_handler,
-        )(block, typeobject)
+        style = getattr(self, f"node_{name}", self._style_handler)(block, typeobject)
         assert len(style) == 3, "Style handler must return (shape, color, label)"
         assert all(
             isinstance(attr, str) or attr or attr is None for attr in style
@@ -279,10 +345,7 @@ class CFG(object):
         return style
 
     def stylize_edge(
-        self,
-        link: Link,
-        default: None = None,
-        default_target: None = None,
+        self, link: Link, default: None = None, default_target: None = None
     ) -> tuple:
         typeobject = link.source.type(default)
         name = typeobject.__name__ if isinstance(typeobject, type) else typeobject
@@ -309,6 +372,7 @@ class CFG(object):
         block: FuncBlock,
         last: str = None,
         visited: Set[Block] = set(),
+        interactive: bool = False,
     ) -> None:
         if block in visited:
             return
@@ -317,22 +381,12 @@ class CFG(object):
         graph.node(
             str(block.id),
             label=block.name,
-            _attributes={
-                "shape": shape,
-                "color": color,
-                "style": "filled",
-            },
+            _attributes={"shape": shape, "color": color, "style": "filled"},
         )
         if last is not None:
-            graph.edge(
-                last,
-                str(block.id),
-                _attributes={
-                    "color": "black",
-                },
-            )
+            graph.edge(last, str(block.id), _attributes={"color": "black"})
         for arg in block.args:
-            self._visit_func(graph, arg, str(block.id))
+            self._visit_func(graph, arg, str(block.id), interactive=interactive)
 
     def _visit_blocks(
         self,
@@ -341,6 +395,7 @@ class CFG(object):
         visited: Set[Block] = set(),
         calls: bool = True,
         format: str = None,
+        interactive: bool = False
     ) -> None:
         # Don't visit blocks twice.
         if block in visited:
@@ -373,12 +428,16 @@ class CFG(object):
             str(block.id),
             label=nodelabel,
             _attributes={
-                "style": "filled",
+                "style": f"filled,{self.border_style(block, interactive)}",
                 "shape": nodeshape,
-                "fillcolor": nodecolor,
+                "fillcolor": self.fillcolor(block, interactive, nodecolor),
             },
         )
 
+        if isinstance(block, TryBlock):
+            for except_block in block.except_blocks.values():
+                self._visit_blocks(graph, except_block, visited, calls, format)
+
         if calls and block.func_calls:
             calls_node = str(block.id) + "_calls"
 
@@ -393,7 +452,7 @@ class CFG(object):
                     "rankdir": "TB",
                     "ranksep": "0.02",
                     "style": "filled",
-                    "color": "purple",
+                    "color": self.fillcolor(block, interactive, "purple"),
                     "compound": "true",
                     "shape": self.node_styles[ast.Call][0],
                     "label": "",
@@ -401,17 +460,18 @@ class CFG(object):
             )
             # Generate control flow edges for function arguments
             for func_block in block.func_blocks:
-                    graph.edge(
-                        str(block.id),
-                        str(func_block.id),
-                        label="calls",
-                        _attributes={"style": "dashed"},
-                    )
-                    self._visit_func(
-                        calls_subgraph,
-                        func_block,
-                        visited=set(),
-                    )
+                graph.edge(
+                    str(block.id),
+                    str(func_block.id),
+                    label="calls",
+                    _attributes={"style": "dashed"},
+                )
+                self._visit_func(
+                    calls_subgraph,
+                    func_block,
+                    visited=set(),
+                    interactive=interactive,
+                )
             graph.subgraph(calls_subgraph)
 
             tmp = ""
@@ -423,9 +483,9 @@ class CFG(object):
                         input_node,
                         label=line,
                         _attributes={
-                            "style": "filled",
+                            "style": f"filled,{self.border_style(block, interactive)}",
                             "shape": nodeshape,
-                            "fillcolor": nodecolor,
+                            "fillcolor": self.fillcolor(block, interactive, nodecolor),
                         },
                     )
                     graph.edge(input_node, str(block.id))  # yellow
@@ -435,11 +495,17 @@ class CFG(object):
                     line += "\l"
                     tmp += line
 
-
         # Recursively visit all the blocks of the CFG.
         for exit in block.exits:
             assert block == exit.source
-            self._visit_blocks(graph, exit.target, visited, calls=calls, format=format)
+            self._visit_blocks(
+                graph,
+                exit.target,
+                visited,
+                calls=calls,
+                format=format,
+                interactive=interactive,
+            )
             edgeshape, edgecolor, edgelabel = self.stylize_edge(exit)
             graph.edge(
                 str(block.id),
@@ -448,39 +514,88 @@ class CFG(object):
                 _attributes={"color": edgecolor},
             )
 
-    def _build_visual(self, format: str = "pdf", calls: bool = True) -> Digraph:
+    subgraphs: dict = Counter()
+
+    def _build_visual(self, format: str = "pdf", calls: bool = True, interactive: bool = False, build_own=False) -> Digraph:
         graph = gv.Digraph(
-            name="cluster" + self.name,
+            name=f"cluster{self.subgraphs[self.name]}{self.name}",
             format=format,
             graph_attr={
                 "label": self.name,
                 "rankdir": "TB",
                 "ranksep": "0.02",
                 "compound": "True",
+                "pack": "False",
             },
         )
+        self.subgraphs[self.name] += 1
+        if self.entryblock is None:
+            raise TypeError("Expected self.entryblock to be not None but type is None")
         self._visit_blocks(
-            graph, self.entryblock, visited=set(), calls=calls, format=format
+            graph,
+            self.entryblock,
+            visited=set(),
+            calls=calls,
+            format=format,
+            interactive=interactive,
         )
+        if build_own:
+            return graph
         # Build the subgraphs for the function definitions in the CFG and add
         # them to the graph.
         for subcfg in self.classcfgs:
             subgraph = self.classcfgs[subcfg]._build_visual(
                 format=format,
                 calls=calls,
+                interactive=interactive,
             )
             graph.subgraph(subgraph)
 
         for subcfg in self.functioncfgs:
             subgraph = self.functioncfgs[subcfg]._build_visual(
-                format=format, calls=calls
+                format=format,
+                calls=calls,
+                interactive=interactive,
             )
             graph.subgraph(subgraph)
         return graph
+    
+
+    @staticmethod
+    def border_style(block: Block, interactive: bool):
+        if interactive:
+            return "bold" if block.outline else "dashed"
+        else:
+            return "solid"
+    
+    @staticmethod
+    def fillcolor(block: Block, interactive: bool, color: str, default: str = ""):
+        if interactive:
+            if block.highlight:
+                return color
+            else:
+                return default
+        else:
+            return color
+
+
+    def build_key(self, **kwargs):
+        """Build key graph"""
+        key = self._build_key_subgraph(kwargs.get("format", "svg"))
+        return key.render(**kwargs)    
 
     def build_visual(
-        self, filepath: str, format: str, calls: bool = True, show: bool = True
-    ) -> None:
+        self,
+        filepath: str,
+        format: str,
+        calls: bool = True,
+        show: bool = True,
+        cleanup=True,
+        directory=None,
+        interactive=False,
+        build_keys=True,
+        build_own=False,
+    ) -> str:
         """
         Build a visualisation of the CFG with graphviz and output it in a DOT
         file.
@@ -492,21 +607,37 @@ class CFG(object):
             show: A boolean indicating whether to automatically open the output
                   file after building the visualisation.
         """
-        graph = self._build_visual(format, calls)
-        graph.subgraph(self._build_key_subgraph(format))
-
-        graph.render(filepath, view=show)
+        if cleanup is None: # Not sure why I did this.
+            cleanup = bool(os.environ["PY2CFG_CLEANUP"])
+        
+        graph = self._build_visual(format, calls, interactive, build_own)
+        if build_keys:
+            graph.subgraph(self._build_key_subgraph(format))
+        return graph.render(filepath, view=show, cleanup=cleanup, directory=directory)
 
     def __iter__(self) -> Iterator[Block]:
         """
         Generator that yields all the blocks in the current graph, then
         recursively yields from any sub graphs
         """
-        visited = set()
-        to_visit: List[Block] = [self.entryblock]
+        yield from self.own_blocks()
+        for subcfg in self.classcfgs.values():
+            yield from subcfg
 
+        for subcfg in self.functioncfgs.values():
+            yield from subcfg
+
+    def own_blocks(self) -> Iterator[Block]:
+        """
+        Generator that yields all blocks in the current graph, excluding any
+        subgraphs
+        """
+        visited = set()
+        if self.entryblock is None:
+            raise TypeError("Expected self.entryblock to be not None but type is None")
+        to_visit: Deque[Block] = deque([self.entryblock])
         while to_visit:
-            block = to_visit.pop(0)
+            block = to_visit.popleft()
             visited.add(block)
             for exit_ in block.exits:
                 if exit_.target in visited or exit_.target in to_visit:
@@ -514,16 +645,16 @@ class CFG(object):
                 to_visit.append(exit_.target)
             yield block
 
-        for subcfg in self.classcfgs.values():
-            yield from subcfg
-
-        for subcfg in self.functioncfgs.values():
-            yield from subcfg
-
     def node_ClassDef(self, block: Block, typeobj: type) -> Tuple[str, str, str]:
         text = block.get_source().splitlines()[0] + "...\n"
         return (*self.node_styles.get(typeobj, self.DEFAULT), text)
 
+    def node_Try(self, block: Block, typeobj: type) -> Tuple[str, str, str]:
+        return (
+            *self.node_styles.get(typeobj, self.DEFAULT),
+            "\n".join(block.get_source().splitlines()),
+        )
+
     def edge_If(self, link: Link) -> Tuple[None, str, str]:
         _, _, nodelabel = self.stylize_node(link.source)
         color = "red"
@@ -548,4 +679,48 @@ class CFG(object):
             color = "green"
         return None, color, edgelabel
 
+    def bsearch(self, lineno: int, lst:Optional[list] = None) -> Optional[Block]:
+        """Search for a block at line"""
 
+        def _bsearch(lst, low, high, line):
+            if high >= low:
+                mid = (low + high) // 2
+                block = lst[mid]
+                if block.at() <= line <= block.end():
+                    return block
+                elif line < block.at():
+                    return _bsearch(lst, low, mid - 1, line)
+                else:
+                    return _bsearch(lst, mid + 1, high, line)
+            else:
+                return None
+
+        lst = list(self) if lst is None else lst  # Already sorted by lineno
+        return _bsearch(lst, 0, len(lst) - 1, lineno)
+
+    outlined_block = Block(0)
+    highlighted_blocks: List[Block] = []
+
+    def outline_block(self, lineno, lst=None) -> int:
+        """In interactive mode, outline the block at lineno"""
+        self.outlined_block.outline = False
+        if (block := self.bsearch(lineno, lst)) is None:
+            return -1
+        self.outlined_block = block
+        block.outline = True
+        return 0
+
+
+    def highlight_blocks(self, lines, blocks=None) -> int:
+        """In interactive mode, highlight the block at lineno"""
+        for block in self.highlighted_blocks:
+            block.highlight = False
+        
+        highlighted_blocks: List[Block] = []
+        for line in list(lines):
+            if (block := self.bsearch(line, blocks)) is not None:
+                block.highlight = True
+                highlighted_blocks.append(block)
+            
+        self.highlighted_blocks = highlighted_blocks
+        return 0
